import sys
import os

cpu_index = 0
cpu_total = 1
job_index = 0

if len(sys.argv)>=3:
	cpu_index = int(sys.argv[1])
	cpu_total = int(sys.argv[2])
else:
	print("warning : arguments might be missing")

def add_job(func):
	global job_index
	if job_index == cpu_index:
		print(f"I'm running job {job_index}")
		func()
	job_index = (job_index+1)%cpu_total

###############################################################

from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import anyact as aa

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from mnist import fmnist,mnist,mnist_perf,mnist_perf_light
import mnist_analysis

import matplotlib.pyplot as plt

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

def expspace(x0,x1,N): return torch.exp(torch.linspace(np.log(x0),np.log(x1),N))

def gradiator(φ,grad_φ):
	class GradiatorFunction(torch.autograd.Function):
		@staticmethod
		def forward(ctx,x, θ):
			y = φ(x, θ)
			ctx.save_for_backward(x,θ,y) # save input for backward pass
			return y
		@staticmethod
		def backward(ctx, grad_o):
			# if not ctx.needs_input_grad[0]: return None # if grad not required, don't compute
			x,θ,y = ctx.saved_tensors # restore input from context
			g_ix, g_iθ = grad_φ(x,θ,y,grad_o)
			return g_ix, g_iθ
	return GradiatorFunction.apply

###################################################################################
###################################################################################
###################################################################################

def bernoulli(shape,p): return torch.bernoulli(torch.ones(shape)*p)

def random_sign_flip(x,proba):
	"""
		proba=0   : never flip
		proba=0.5 : complete random sign
		proba=1.0 : always flip
	"""
	flip = bernoulli(x.shape,proba).to(x.device)
	return x*(1.0-2.0*flip)

def only_sign(x): return 1.0-2.0*(x<0)

class BrokenLinear(nn.Module):
	def __init__(self,dim_i,dim_o,
		θ_noise_static_σ=0.0,
		θ_noise_static_μ=0.0,
		# gθ_noise_static_σ=0.0,
		# gθ_noise_static_μ=0.0,
		# gx_noise_static_σ=0.0,
		# gx_noise_static_μ=0.0,
		gθ_flip_static=0.0,
		gx_flip_static=0.0,
		gθ_flip_dynamic=0.0, # 0.48
		gx_flip_dynamic=0.0, # 0.48
		# θ_only_sign=True ,θ_noise_dynamic=0.00, θ_flip=0.3,
		gθ_only_sign=False,
		gx_only_sign=False,
	):
		super().__init__()
		self.θ = nn.parameter.Parameter(torch.randn(dim_i,dim_o))
		self.b = nn.parameter.Parameter(torch.randn(dim_o))
		self.θ_noise_static  = nn.parameter.Parameter( θ_noise_static_μ +  θ_noise_static_σ*torch.randn(dim_i,dim_o),False)
		# self.gθ_noise_static = nn.parameter.Parameter(gθ_noise_static_μ + gθ_noise_static_σ*torch.randn(dim_i,dim_o),False)
		# self.gx_noise_static = nn.parameter.Parameter(gx_noise_static_μ + gx_noise_static_σ*torch.randn(dim_i      ),False)
		self.gθ_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i,dim_o),gθ_flip_static), False)
		self.gx_flip_static = nn.parameter.Parameter( random_sign_flip(torch.ones(dim_i      ),gx_flip_static), False)
		def fwd(x,θ):
			return torch.einsum('ij,ni->nj', θ, x) + self.b
		def bwd(x,θ,y,g_o,debug=False):
			############## break θ
			# print(torch.sum(torch.abs(θ)))
			# print(torch.sum(torch.abs(self.θ_noise_static)))
			θ = θ + self.θ_noise_static
			# θ = self.θ_noise_static
			# if θ_only_sign: θ = only_sign(θ)
			# θ = random_sign_flip(θ, θ_flip)
			############## normal stuff
			g_iθ = torch.einsum('ni,nj->nij', x, g_o) # produit extérieur
			g_ix = torch.einsum('nj,ij->ni' , g_o, θ) # θ⋅x
			############## static breaking
			# g_iθ = g_iθ + self.gθ_noise_static
			# g_ix = g_ix + self.gx_noise_static
			g_iθ = g_iθ * self.gθ_flip_static
			g_ix = g_ix * self.gx_flip_static
			############## dynamic breaking
			if gx_only_sign: g_ix = only_sign(g_ix) # this seems fine...
			if gθ_only_sign: g_iθ = only_sign(g_iθ) # this is bad !
			if not gx_flip_dynamic==0.0 : g_ix = random_sign_flip(g_ix, gx_flip_dynamic)  # this is
			if not gθ_flip_dynamic==0.0 : g_iθ = random_sign_flip(g_iθ, gθ_flip_dynamic)  # but this is ok ... ??
			############## output
			return g_ix, g_iθ
		self.model = aa.gradiator(fwd,bwd)
	def forward(self,x): return self.model(x,self.θ)

# L = BrokenLinear(10,20)
# x = torch.randn(5,10)
# L(x).shape

class BrokenClassifier(nn.Module):
	def __init__(self,layers,gen_linear):
		super().__init__()
		layers = [int(np.round(x)) for x in layers]
		self.L  = [gen_linear(λ1,λ2)  for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.BN = [nn.BatchNorm1d(λ2) for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.σ  = [nn.GELU()          for (λ1,λ2) in zip(layers[:-1],layers[1:])]
		self.model = nn.Sequential(*reduce(operator.add, zip(self.L,self.σ,self.BN))[:-2])
		self.softmax = nn.Softmax(dim=1)
	def forward(self,x):
		return self.softmax(self.model(nn.Flatten()(x)))

# layers = [28*28,32,10]
# cl = BrokenClassifier(layers)
# cl(torch.randn(10,1,28,28))

# def gen_linear(λ1,λ2): return BrokenLinear(λ1,λ2) # TODOOOOOOOOOOOOO
# layers = [28*28,32,10]
# cl = BrokenClassifier(layers,gen_linear)
# cl(torch.randn(10,1,28,28))

fmnist_L_x = mnist_analysis.fmnist_L_x
fmnist_L_y = mnist_analysis.fmnist_L_y
fmnist_T_x = mnist_analysis.fmnist_T_x
fmnist_T_y = mnist_analysis.fmnist_T_y

os.makedirs("models",exist_ok=True)

def analysis(
	N,
	θ_noise_static_σ=0.0,
	θ_noise_static_μ=0.0,
	gθ_flip_static=0.0,
	gx_flip_static=0.0,
	gθ_flip_dynamic=0.0, # 0.48
	gx_flip_dynamic=0.0, # 0.48
	gθ_only_sign=False,
	gx_only_sign=False,
):
	identifier = f"fc({N})-θnoise=[{θ_noise_static_σ},{θ_noise_static_μ}]-flip_static=[{gθ_flip_static},{gx_flip_static}]"
	identifier = f"{identifier}-flip_dyn=[{gθ_flip_dynamic},{gx_flip_dynamic}]-sgn_only=[{gx_only_sign},{gθ_only_sign}]"
	# model_path = f"models/{identifier}"
	def gen_linear(λ1,λ2):
		return BrokenLinear(λ1,λ2,
			θ_noise_static_σ = θ_noise_static_σ ,
			θ_noise_static_μ = θ_noise_static_μ ,
			gθ_flip_static   = gθ_flip_static   ,
			gx_flip_static   = gx_flip_static   ,
			gθ_flip_dynamic  = gθ_flip_dynamic  ,
			gx_flip_dynamic  = gx_flip_dynamic  ,
			gθ_only_sign     = gθ_only_sign     ,
			gx_only_sign     = gx_only_sign     ,
		)
	classifiers = BrokenClassifier(expspace(28*28,10,N+1),gen_linear)
	mnist_analysis.analysis_classification(classifiers, identifier, nn.MSELoss(),batch_size=2048,epochs=16)
	# torch.save(classifiers, model_path)

# analysis(4,0.0,0.0)
# analysis(4,0.5,0.0)
# analysis(4,20000,0.0)
analysis(4,1,0.0)
# analysis(4,0.0,200)


# for N in [2,4]:
#  for  θ_noise_static_σ in [0,0.5,1.0]
#   for θ_noise_static_μ in [0,0.5,1.0]
#    add_job(lambda:analysis(
#      N,
#      θ_noise_static_σ,
#      θ_noise_static_μ,
#    ))


# 	for gθ_flip_static   = gθ_flip_static   ,
# 	for gx_flip_static   = gx_flip_static   ,
# 	for gθ_flip_dynamic  = gθ_flip_dynamic  ,
# 	for gx_flip_dynamic  = gx_flip_dynamic  ,
# 	for gθ_only_sign     = gθ_only_sign     ,
# 	for gx_only_sign     = gx_only_sign     ,

# 								add_job(lambda:analysis(
# 									N,
# 									θ_noise_static_σ,
# 									θ_noise_static_μ,
# 									gθ_flip_static,
# 									gx_flip_static,
# 									gθ_flip_dynamic,
# 									gx_flip_dynamic,
# 									gθ_only_sign,
# 									gx_only_sign,
# 								))
